'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends2 = require('babel-runtime/helpers/extends');

var _extends3 = _interopRequireDefault(_extends2);

var _defineProperty2 = require('babel-runtime/helpers/defineProperty');

var _defineProperty3 = _interopRequireDefault(_defineProperty2);

var _slicedToArray2 = require('babel-runtime/helpers/slicedToArray');

var _slicedToArray3 = _interopRequireDefault(_slicedToArray2);

var _regenerator = require('babel-runtime/regenerator');

var _regenerator2 = _interopRequireDefault(_regenerator);

var _promise = require('babel-runtime/core-js/promise');

var _promise2 = _interopRequireDefault(_promise);

var _asyncToGenerator2 = require('babel-runtime/helpers/asyncToGenerator');

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

var _getIterator2 = require('babel-runtime/core-js/get-iterator');

var _getIterator3 = _interopRequireDefault(_getIterator2);

var _getPrototypeOf = require('babel-runtime/core-js/object/get-prototype-of');

var _getPrototypeOf2 = _interopRequireDefault(_getPrototypeOf);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

var _possibleConstructorReturn2 = require('babel-runtime/helpers/possibleConstructorReturn');

var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);

var _inherits2 = require('babel-runtime/helpers/inherits');

var _inherits3 = _interopRequireDefault(_inherits2);

exports.default = subscribe;
exports.isModelKey = isModelKey;
exports.customClone = customClone;

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _hoistNonReactStatics = require('hoist-non-react-statics');

var _hoistNonReactStatics2 = _interopRequireDefault(_hoistNonReactStatics);

var _Doc = require('./types/Doc');

var _Doc2 = _interopRequireDefault(_Doc);

var _Query = require('./types/Query');

var _Query2 = _interopRequireDefault(_Query);

var _QueryExtra = require('./types/QueryExtra');

var _QueryExtra2 = _interopRequireDefault(_QueryExtra);

var _Local = require('./types/Local');

var _Local2 = _interopRequireDefault(_Local);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// TODO: Explore the possibilities to optimize _.isEqual and _.clone
// http://stackoverflow.com/q/122102

// Deprecate the reactiveProps list of items.
// Make all queries reactive by default (evaluate subscriptions getter
// whenever props change).
var REACTIVE_BY_DEFAULT_OVERRIDE = typeof window !== 'undefined' && window.__sharedbReactiveByDefault;
var REACTIVE_BY_DEFAULT = REACTIVE_BY_DEFAULT_OVERRIDE != null ? REACTIVE_BY_DEFAULT_OVERRIDE : true;

/**
 * ShareDB subscriptions decorator.
 * @param [reactiveProp] {...string} - DEPRECATED! names of the props which trigger resubscribe
 *    Update subscription when any of them changes.
 * @param getSubscriptions {Function} - (props, state) Retrieve initial subscriptions data
 * @returns {Function}
 * @constructor
 * @example
 * |  @subscribe('toContentIds', 'fromContentIds', (props) => {
 * |    return {
 * |      template: ['templates', props.templateId],
 * |      sections: ['sections', { instructions: true, $sort: { createdAt: -1 } }],
 * |      toTexts: ['texts', { _id: { $in: props.toContentIds } }],
 * |      fromTexts: ['texts', { _id: { $in: props.fromContentIds } }]
 * |    }
 * |  })
 * @example Extra query (counting amount of documents)
 * |  @subscribe((props) => ({
 * |    contentsCount: ['contents', {
 * |      $count: true,
 * |      sectionId: props.sectionId
 * |    }]
 * |  }))
 */
function subscribe() {
  var getSubscriptions = arguments[arguments.length - 1];
  if (typeof getSubscriptions !== 'function') {
    throw new Error('[@subscribe] last argument (getSubscriptions) must be a function.');
  }
  var reactiveProps = Array.prototype.slice.call(arguments, 0, arguments.length - 1);
  if (reactiveProps.some(function (i) {
    return typeof i !== 'string';
  })) {
    throw new Error('[@subscribe] reactiveProps must be strings.');
  }
  return function decorateTarget(DecoratedComponent) {
    var SubscriptionsContainer = function (_React$Component) {
      (0, _inherits3.default)(SubscriptionsContainer, _React$Component);

      function SubscriptionsContainer(props) {
        (0, _classCallCheck3.default)(this, SubscriptionsContainer);

        var _this = (0, _possibleConstructorReturn3.default)(this, (SubscriptionsContainer.__proto__ || (0, _getPrototypeOf2.default)(SubscriptionsContainer)).call(this, props));

        _this.subscriptions = _this.getCurrentSubscriptions(props);
        _this.state = {};
        _this.listeners = {};
        _this.init();
        return _this;
      }

      (0, _createClass3.default)(SubscriptionsContainer, [{
        key: 'componentWillUnmount',
        value: function componentWillUnmount() {
          this.unmounted = true;
          for (var key in this.items) {
            this.destroyItem(key);
          }
          delete this.items;
        }

        // Update only after everything loaded

      }, {
        key: 'shouldComponentUpdate',
        value: function shouldComponentUpdate(nextProps, nextState) {
          return !!this.loaded;
        }
      }, {
        key: 'getCurrentSubscriptions',
        value: function getCurrentSubscriptions(props) {
          return getSubscriptions && getSubscriptions(props) || {};
        }

        // Update queries whenever props change and if the subscriptions
        // object changes as a result of that

      }, {
        key: 'componentWillReceiveProps',
        value: function componentWillReceiveProps(nextProps) {
          var _this2 = this;

          if (!REACTIVE_BY_DEFAULT) {
            var updateQueries = reactiveProps.some(function (reactiveProp) {
              return !_lodash2.default.isEqual(_this2.props[reactiveProp], nextProps[reactiveProp]);
            });
            if (!updateQueries) return;
          }
          var prevSubscriptions = this.subscriptions;
          this.subscriptions = this.getCurrentSubscriptions(nextProps);
          var keys = _lodash2.default.union(_lodash2.default.keys(prevSubscriptions), _lodash2.default.keys(this.subscriptions));
          keys = _lodash2.default.uniq(keys);
          var _iteratorNormalCompletion = true;
          var _didIteratorError = false;
          var _iteratorError = undefined;

          try {
            for (var _iterator = (0, _getIterator3.default)(keys), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
              var key = _step.value;

              if (!_lodash2.default.isEqual(this.subscriptions[key], prevSubscriptions[key])) {
                if (prevSubscriptions[key]) this.destroyItem(key);
                if (this.subscriptions[key]) {
                  this.initItem(key);
                } else {
                  // If the subscription was there before but now
                  // it's gone, we should remove the item's data
                  this.removeItemData(key);
                }
              }
            }
          } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion && _iterator.return) {
                _iterator.return();
              }
            } finally {
              if (_didIteratorError) {
                throw _iteratorError;
              }
            }
          }
        }
      }, {
        key: '_isExtraQuery',
        value: function _isExtraQuery(queryParams) {
          return queryParams.$count || queryParams.$aggregate;
        }
      }, {
        key: 'init',
        value: function () {
          var _ref = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee() {
            var key, _constructor, _key;

            return _regenerator2.default.wrap(function _callee$(_context) {
              while (1) {
                switch (_context.prev = _context.next) {
                  case 0:
                    this.items = {};
                    this.itemKeys = {};
                    for (key in this.subscriptions) {
                      _constructor = this.getItemConstructor(this.subscriptions[key]);

                      this.items[key] = new _constructor(key, this.subscriptions[key]);
                    }
                    // Init all items
                    _context.next = 5;
                    return _promise2.default.all(_lodash2.default.map(this.items, function (i) {
                      return i.init();
                    }));

                  case 5:
                    if (!this.unmounted) {
                      _context.next = 7;
                      break;
                    }

                    return _context.abrupt('return');

                  case 7:
                    this.updateAllData();
                    // Start listening for updates
                    for (_key in this.items) {
                      this.listenForItemUpdates(_key);
                    }
                    this.loaded = true;
                    this.forceUpdate();

                  case 11:
                  case 'end':
                    return _context.stop();
                }
              }
            }, _callee, this);
          }));

          function init() {
            return _ref.apply(this, arguments);
          }

          return init;
        }()
      }, {
        key: 'initItem',
        value: function () {
          var _ref2 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee2(key) {
            var constructor;
            return _regenerator2.default.wrap(function _callee2$(_context2) {
              while (1) {
                switch (_context2.prev = _context2.next) {
                  case 0:
                    constructor = this.getItemConstructor(this.subscriptions[key]);

                    this.items[key] = new constructor(key, this.subscriptions[key]);
                    _context2.next = 4;
                    return this.items[key].init();

                  case 4:
                    if (!this.unmounted) {
                      _context2.next = 6;
                      break;
                    }

                    return _context2.abrupt('return');

                  case 6:
                    this.updateItemData(key);
                    this.listenForItemUpdates(key);

                  case 8:
                  case 'end':
                    return _context2.stop();
                }
              }
            }, _callee2, this);
          }));

          function initItem(_x) {
            return _ref2.apply(this, arguments);
          }

          return initItem;
        }()
      }, {
        key: 'destroyItem',
        value: function destroyItem(key) {
          this.items[key].destroy();
          delete this.items[key];
        }
      }, {
        key: 'getItemConstructor',
        value: function getItemConstructor(subscription) {
          if (typeof subscription === 'string') return _Local2.default;

          var _subscription = (0, _slicedToArray3.default)(subscription, 2),
              params = _subscription[1];

          return typeof params === 'string' || !params ? _Doc2.default : this._isExtraQuery(params) ? _QueryExtra2.default : _Query2.default;
        }
      }, {
        key: 'listenForItemUpdates',
        value: function listenForItemUpdates(key) {
          this.items[key].on('update', this.updateItemData.bind(this, key));
        }
      }, {
        key: 'updateAllData',
        value: function updateAllData() {
          var _this3 = this;

          var data = {};
          _lodash2.default.reduce(this.items, function (data, item, key) {
            var itemData = item.getData();
            _this3.itemKeys[key] = _lodash2.default.keys(itemData);
            return _lodash2.default.merge(data, itemData);
          }, data);
          this.setState(customClone(data));
        }

        // Update item data and also remove any obsolete data
        // (old keys are tracked it itemKeys)

      }, {
        key: 'updateItemData',
        value: function updateItemData(key) {
          var data = this.items[key].getData();
          var oldItemKeys = this.itemKeys[key] || [];
          var itemKeys = _lodash2.default.keys(data);
          this.itemKeys[key] = itemKeys;
          var equal = this.items[key].isEqual(data, _lodash2.default.pick(this.state, oldItemKeys));
          if (equal) return;
          var removeValues = {};
          var _iteratorNormalCompletion2 = true;
          var _didIteratorError2 = false;
          var _iteratorError2 = undefined;

          try {
            for (var _iterator2 = (0, _getIterator3.default)(_lodash2.default.difference(oldItemKeys, itemKeys)), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
              var itemKey = _step2.value;

              _lodash2.default.merge(removeValues, (0, _defineProperty3.default)({}, itemKey, null));
            }
            // TODO: remove log
            // console.log('--UPDATE', this.state, data)
          } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion2 && _iterator2.return) {
                _iterator2.return();
              }
            } finally {
              if (_didIteratorError2) {
                throw _iteratorError2;
              }
            }
          }

          this.setState(_lodash2.default.merge(removeValues, customClone(data)));
        }
      }, {
        key: 'removeItemData',
        value: function removeItemData(key) {
          var oldItemKeys = this.itemKeys[key] || [];
          if (oldItemKeys.length === 0) return;
          var removeValues = {};
          var _iteratorNormalCompletion3 = true;
          var _didIteratorError3 = false;
          var _iteratorError3 = undefined;

          try {
            for (var _iterator3 = (0, _getIterator3.default)(oldItemKeys), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
              var itemKey = _step3.value;

              _lodash2.default.merge(removeValues, (0, _defineProperty3.default)({}, itemKey, null));
            }
          } catch (err) {
            _didIteratorError3 = true;
            _iteratorError3 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion3 && _iterator3.return) {
                _iterator3.return();
              }
            } finally {
              if (_didIteratorError3) {
                throw _iteratorError3;
              }
            }
          }

          delete this.itemKeys[key];
          this.setState(removeValues);
        }
      }, {
        key: 'render',
        value: function render() {
          if (this.loaded) {
            return _react2.default.createElement(DecoratedComponent, (0, _extends3.default)({}, this.props, this.state));
          } else {
            // When in React Native env, don't use any loading spinner
            if (typeof navigator !== 'undefined' && navigator.product === 'ReactNative') {
              return null;
            } else {
              return _react2.default.createElement('div', { className: 'Loading' });
            }
          }
        }
      }]);
      return SubscriptionsContainer;
    }(_react2.default.Component);

    return (0, _hoistNonReactStatics2.default)(SubscriptionsContainer, DecoratedComponent);
  };
}

function isModelKey(key) {
  return (/^\$/.test(key)
  );
}

// Don't clone models (which are starting with $)
function customClone(data) {
  var res = {};
  for (var key in data) {
    if (isModelKey(key)) {
      res[key] = data[key];
    } else {
      res[key] = _lodash2.default.cloneDeep(data[key]);
    }
  }
  return res;
}